/*
** Module   : P2JField.java
** Abstract : Helper class for dynamic temporary tables.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SVL 20130624 Created initial version.
** 002 SVL 20130822 Continued implementation.
*/

package com.goldencode.p2j.persist;

import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.compile.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.xml.*;
import org.apache.commons.logging.*;
import org.hibernate.cfg.*;
import org.hibernate.internal.*;
import org.hibernate.mapping.*;
import java.io.*;
import java.util.*;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.*;

/**
 * Helper class for dynamic temporary tables.
 */
public class DynamicTablesHelper
{
   /** Base package for generated DMOs. */
   public final static String DMO_BASE_PACKAGE = "com.goldencode.p2j.persist.dynamic";

   /** Category name for DMO implementation classes, used in conversion rules. */
   private final static String DMO_IMPL_CATEGORY = "impl";

   /** Category name for DMO interfaces, used in conversion rules. */
   private final static String DMO_IFACE_CATEGORY = "iface";

   /** Logger. */
   private final static Log LOG = LogFactory.getLog(DynamicTablesHelper.class);

   /** Singleton instance. */
   private static volatile DynamicTablesHelper instance;

   /** Param type to Progress type. */
   private final static Map<ParmType, Integer> propTypeToProgress
                                                = new HashMap<ParmType, Integer>();

   /** Containers for storing context-specific data. */
   private final static ContextLocal<Context> context = new ContextLocal<Context>()
   {
      public Context initialValue()
      {
         return new Context();
      }
   };

   /** Counter used for generation of unique names for dynamic tables. */
   private final AtomicLong tableCounter = new AtomicLong(1);

   /** Shared Hibernate <code>Configuration</code> object. */
   private final Configuration cfg =
               DatabaseManager.getConfigBySchema(DatabaseManager.TEMP_TABLE_SCHEMA);

   static
   {
      propTypeToProgress.put(ParmType.CHAR, ProgressParserTokenTypes.FIELD_CHAR);
      propTypeToProgress.put(ParmType.INT, ProgressParserTokenTypes.FIELD_INT);
      propTypeToProgress.put(ParmType.INT64, ProgressParserTokenTypes.FIELD_INT64);
      propTypeToProgress.put(ParmType.DEC, ProgressParserTokenTypes.FIELD_DEC);
      propTypeToProgress.put(ParmType.LOG, ProgressParserTokenTypes.FIELD_LOGICAL);
      propTypeToProgress.put(ParmType.DATE, ProgressParserTokenTypes.FIELD_DATE);
      propTypeToProgress.put(ParmType.DT, ProgressParserTokenTypes.FIELD_DATETIME);
      propTypeToProgress.put(ParmType.DTTZ, ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      propTypeToProgress.put(ParmType.HANDLE, ProgressParserTokenTypes.FIELD_HANDLE);
      propTypeToProgress.put(ParmType.RECID, ProgressParserTokenTypes.FIELD_RECID);
      propTypeToProgress.put(ParmType.ROWID, ProgressParserTokenTypes.FIELD_ROWID);
   }

   /**
    * Get singleton instance of the helper.
    *
    * @return  singleton instance of the helper.
    */
   public static DynamicTablesHelper getInstance()
   {
      if (instance == null)
      {
         synchronized (DynamicTablesHelper.class)
         {
            if (instance == null)
            {
               instance = new DynamicTablesHelper();
               PatternEngine.setDebugLevel(ConversionDriver.MSG_NONE);
            }
         }
      }

      return instance;
   }

   /**
    * Get Hibernate field name by 4GL field name for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  name4gl
    *         4GL field name.
    *
    * @return Hibernate field name for the given DMO class and 4GL field name or <code>null</code>
    *         if no corresponding name was found.
    */
   public static String getHibernateFieldName(Class dmoIface, String name4gl)
   {
      Map<String, String> map = context.get().fields4GLtoHibernate.get(dmoIface);
      if (map == null)
         return null;

      return map.get(name4gl);
   }

   /**
    * Generates DMO interface, implementation class and mapping document. TODO
    *
    * @param builder
    *        Temporary table builder containing definitions of the target table.
    *
    * @return <code>true</code> on success.
    */
   public boolean generateDynamicDMO(TempTableBuilder builder)
   {
      long tableId = getUniqueTableId();

      String tableName = String.format("dtt%d", tableId);
      String ifaceName = String.format("DynamicRecord%d", tableId);
      String implClassName = String.format("DynamicRecord%dImpl", tableId);
      Map<String, String> clsFields4GLtoHibernate = new HashMap<String, String>();

      long fieldId = 1;
      for (P2JField field : builder.getFields())
      {
         String fieldNameHibernate = String.format("field%d", fieldId++);
         clsFields4GLtoHibernate.put(field.getName(), fieldNameHibernate);
      }

      Aast root = instance.generateDMOAst(builder, ifaceName, tableName, clsFields4GLtoHibernate);

      try
      {
         PatternEngine engine;
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/p2o");

         root = (Aast) engine.getStoredObject(null);
         root.putAnnotation("permanent", false);
         root.putAnnotation("package", DMO_BASE_PACKAGE);

         // generate .hbm.xml
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/hibernate");
         XmlAst hbmXml = (XmlAst) engine.getStoredObject(ifaceName);

         // generate interface and implementation Java ASTs
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/java_dmo");
         Aast impl = (Aast) engine.getStoredObject(DMO_IMPL_CATEGORY, ifaceName);
         Aast iface = (Aast) engine.getStoredObject(DMO_IFACE_CATEGORY, ifaceName);

         // generate Java code from Java ASTs
         engine = new PatternEngine();
         engine.addTargetAst(impl);
         engine.addTargetAst(iface);
         engine.run("convert/brew");

         ByteArrayOutputStream implCodeStream =
                              (ByteArrayOutputStream) engine.getStoredObject(implClassName);
         ByteArrayOutputStream ifaceCodeStream =
                              (ByteArrayOutputStream) engine.getStoredObject(ifaceName);

         String fullIfaceName = String.format("%s.%s.%s",
                                              DMO_BASE_PACKAGE,
                                              DatabaseManager.TEMP_TABLE_SCHEMA,
                                              ifaceName);
         String fullImplClassName = String.format("%s.%s.impl.%s",
                                                  DMO_BASE_PACKAGE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA,
                                                  implClassName);
         StringBuilder sb = new StringBuilder();
         Map<String, Class<?>> classes = InMemoryCompiler.compile(new String[]
                                                                  {
                                                                     fullIfaceName,
                                                                     fullImplClassName
                                                                  },
                                                                  new String[]
                                                                  {
                                                                      ifaceCodeStream.toString(),
                                                                      implCodeStream.toString()
                                                                  },
                                                                  sb, true);

         if (classes == null)
         {
            LOG.error(sb.toString());
            ErrorManager.displayError("Cannot compile Hibernate entities.");
            return false;
         }

         Class dmoIface = classes.get(fullIfaceName);
         Class dmoClass = classes.get(fullImplClassName);

         MutableSessionFactoryImpl sessionFactory =
               (MutableSessionFactoryImpl) DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
         PersistentClass persistentClass;
         synchronized (cfg)
         {
            // SVL: as far as I've checked Configuration object is not accessed by Hibernate
            // outside synchronized blocks in this class.  DBUtils.makeTypeArray (which also
            // uses Configuration object), accesses Configuration.typeResolver which remains
            // intact in the synchronized blocks.

            cfg.addDocument(XmlAst.astToDom(hbmXml));
            cfg.buildMappings();
            persistentClass = cfg.getClassMapping(dmoClass.getName());
            sessionFactory.add(persistentClass, cfg);
         }

         DMOIndex.registerDynamicTable(dmoIface, dmoClass, builder.getIndexes().values());
         DatabaseManager.registerDynamicTable(dmoIface, persistentClass);

         Map<Class, Map<String, String>> fields = context.get().fields4GLtoHibernate;
         fields.put(dmoIface, clsFields4GLtoHibernate);
         context.get().persistentClasses.put(dmoIface, persistentClass);

         builder.setDmoIface(dmoIface);
         builder.setPersistentClass(persistentClass);
      }
      catch (Exception e)
      {
         LOG.error(e);
         ErrorManager.displayError("Error while generating dynamic table: " + e.getMessage());
         return false;
      }

      return true;
   }

   /**
    * Remove information about the specified dynamic DMO class from Hibernate Configuration,
    * session factory and P2J facilities.
    *
    * @param dmoIface
    *        DMO interface which corresponds the target dynamic DMO class.
    * @param cleanContextLocal
    *        If <code>true</code> then context-local information in P2J facilities will be cleaned
    *        (if may want not do that if the session is closing because all context-local
    *        information will be dropped anyway).
    */
   public void dropDynamicDMO(Class dmoIface,
                              boolean cleanContextLocal)
   {
      Context local = context.get();
      PersistentClass persistentClass = local.persistentClasses.get(dmoIface);

      MutableSessionFactoryImpl sessionFactory =
               (MutableSessionFactoryImpl) DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
      sessionFactory.remove(persistentClass);

      synchronized (cfg)
      {
         cfg.remove(persistentClass);
      }

      //TODO clean InMemoryClassLoader.loaded

      if (cleanContextLocal)
      {
         Class dmoClass = persistentClass.getMappedClass();
         local.persistentClasses.remove(dmoIface);
         local.fields4GLtoHibernate.remove(dmoIface);
         DMOIndex.deregisterDynamicTable(dmoIface, dmoClass);
         DatabaseManager.deregisterDynamicTable(dmoIface, persistentClass);
      }
   }

   /**
    * Create and register dynamic record buffer with the specified name for the specified DMO
    * class.
    *
    * @param  dmoIface
    *         DMO interface which defines DMO's public API.
    * @param  bufferName
    *         Buffer name.
    * @param  undoable
    *         <code>true</code> if changes made to records managed by this buffer can be undone.
    *
    * @return dynamic record buffer for the specified DMO class.
    */
   public Temporary createDynamicBuffer(Class dmoIface,
                                        String bufferName,
                                        boolean undoable)
   {
      Temporary buffer = TemporaryBuffer.define(dmoIface.getDeclaredClasses()[0],
                                                bufferName,
                                                true,
                                                undoable);
      ((TemporaryBuffer) RecordBuffer.get(buffer)).setDynamic();
      TemporaryBuffer.openFreeScope(buffer);
      context.get().dynamicBuffers.add(buffer);

      return buffer;
   }

   /**
    * Generate and return a new unique id.
    *
    * @return  a new unique id.
    */
   long getUniqueTableId()
   {
      return tableCounter.getAndIncrement();
   }

   /**
    * Generate AST that represents the target table and has the same structure as .p2o files from
    * the conversion stage.
    *
    * @param  builder
    *         Temporary table builder containing definitions of the target table.
    * @param  ifaceName
    *         The name of the target DMO interface.
    * @param  tableName
    *         The name of the target SQL table.
    * @param  fields4GLtoHibernate
    *         Map 4GL field names to Hibernate names.
    *
    * @return the root node of the generated AST.
    */
   private Aast generateDMOAst(TempTableBuilder builder,
                               String ifaceName,
                               String tableName,
                               Map<String, String> fields4GLtoHibernate)
   {
      Aast root = new ProgressAst(new CommonToken(ProgressParserTokenTypes.DATABASE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA));

      ProgressAst cls = new ProgressAst(new CommonToken(ProgressParserTokenTypes.TABLE,
                                                        tableName));
      cls.putAnnotation("class-name", ifaceName);
      cls.putAnnotation("table-name", tableName);
      root.addChild(cls);

      for (P2JField field : builder.getFields())
      {
         String fieldName = fields4GLtoHibernate.get(field.getName());
         ParmType type = field.getType();
         int progressType = propTypeToProgress.get(type);
         ProgressAst property = new ProgressAst(new CommonToken(progressType, fieldName));
         property.putAnnotation("format", field.getFormat());
         property.putAnnotation("label", field.getLabel());

         long extent = field.getExtent();
         if (extent > 1)
            property.putAnnotation("extent", extent);

         cls.addChild(property);

         BaseDataType initial = field.getInitial();
         if (initial != null)
         {
            ProgressAst initialNode = new ProgressAst(
                                      new CommonToken(ProgressParserTokenTypes.KW_INIT, ""));
            property.addChild(initialNode);

            int valType;
            if (initial.isUnknown())
               valType = ProgressParserTokenTypes.UNKNOWN_VAL;
            else switch (type)
            {
               case CHAR:
                  valType = ProgressParserTokenTypes.STRING;
                  break;
               case INT:
                  valType = ProgressParserTokenTypes.NUM_LITERAL;
                  break;
               case DEC:
                  valType = ProgressParserTokenTypes.DEC_LITERAL;
                  break;
               case LOG:
                  valType = initial.equals(new logical(true)) ?
                            ProgressParserTokenTypes.BOOL_TRUE :
                            ProgressParserTokenTypes.BOOL_FALSE;
                  break;
               default:
                  throw new IllegalStateException();
            }

            String valStr;
            switch (type)
            {
               case CHAR:
                  valStr = new StringBuilder().append("\"").
                                               append(initial.toStringMessage()).
                                               append("\"").toString();
                  break;
               default:
                  valStr = initial.toStringMessage();
            }

            ProgressAst initialVal = new ProgressAst(
                                        new CommonToken(valType, valStr));
            initialNode.addChild(initialVal);
         }
      }

      assignIds(root, 1);

      return root;
   }

   /**
    * Traverse an AST and assign some unique ids to all of its nodes.
    *
    * @param  root
    *         The root node of the AST.
    * @param  startId
    *         First available id for assignment.
    *
    * @return new first available id for assignment.
    */
   private long assignIds(Aast root, long startId)
   {
      root.setId(startId++);

      int childCount = root.getNumImmediateChildren();
      if (childCount > 0)
      {
         for (int i = 0; i < childCount; i++)
         {
            Aast child = root.getChildAt(i);
            startId = assignIds(child, startId);
         }
      }
      return startId;
   }

   /**
    * Container for storing context-local information about dynamic tables and buffers.
    */
   private static class Context
   implements Finalizable
   {
      /**
       * Mapping of actual field names specified in ADD-NEW-FIELD to Hibernate names ("field1",
       * "field2", ...). DMOs represented by corresponding interface class names.
       */
      private final Map<Class, Map<String, String>> fields4GLtoHibernate =
                                             new HashMap<Class, Map<String, String>>();

      /**
       * Mapping of DMO interfaces to corresponding persistent classes.
       */
      private final Map<Class, PersistentClass> persistentClasses =
                                             new HashMap<Class, PersistentClass>();

      /**
       * Set of all dynamic record buffers created in this context.
       */
      private final Set<Temporary> dynamicBuffers = new HashSet<Temporary>();

      private Context()
      {
         //TransactionManager.registerFinalizable(this, true);
      }

      public void finished()
      {
         /*DynamicTablesHelper instance = getInstance();

         for (Temporary buffer : dynamicBuffers)
         {
            instance.dropDynamicBuffer((Buffer) buffer);
         }

         for (Class dmoIface : persistentClasses.keySet())
         {
            instance.dropDynamicDMO(dmoIface, false);
         } */
      }

      public void iterate()
      {
      }

      public void retry()
      {
      }
   }
}
