/*
** Module   : TempTableBuilder.java
** Abstract : Implementation of the dynamic features for constructing temp-tables  
*              (attributes, methods and constructors).
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 OM  20130129 Created initial version.
** 002 VMN 20130217 Added support for handle-based attributes:
**                  ERROR (KW_ERROR)
** 003 CA  20130221 Added ADM-DATA, UNIQUE-ID, HandleChain and Deletable support.
** 004 CA  20130311 Added missing ADD-NEW-FIELD and ADD-NEW-INDEX API versions, using wrapper 
**                  types. Note that the conversion will use only the APIs with the wrapper type
**                  parameters; the other APIs (using non-wrapper parameters) are removed.
** 005 CS  20130322 Made the class non-abstract and added missing method stubs. 
** 006 CA  20130603 resourceDelete must return a boolean value.
** 006 SVL 20130624 Some methods were implemented.
** 007 SVL 20130822 Some methods and attributes were implemented.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;
import org.hibernate.*;
import org.hibernate.mapping.PersistentClass;

import java.util.*;

/**
 * Implementation of buffer of a dynamic TEMP-TABLE and the associated 
 * attributes and methods.
 * This {@link TempTable} interface is needed by {@link CommonHandle}
 * in order to store the handle to a the table.
 * 
 * @author om
 */
public class TempTableBuilder
extends HandleChain
implements TempTable,
           Deletable
{
   /** datetime-tz 4GL type name. */
   private final static String DATETIMETZ_TYPE_NAME = "datetime-tz";

   /** Special characters (not letters or digits) allowed in dynamic table names. */
   private final static Set<Character> allowedSpecialChars = new HashSet<Character>();

   static
   {
      allowedSpecialChars.addAll(Arrays.asList('-', '#', '$', '%', '&', '_'));
   }

   /**
    * Used during data type name validation. Map that stores the minimal possible length of a data
    * type name abbreviation. E.g. instead "integer" you can specify "integ", "int" or "i" because
    * the minimal length for this type is 1.
    */
   private final static Map<String, Integer> datatypesMatchLen =
                                                new LinkedHashMap<String, Integer>();

   /**
    * <code>true</code> if changes made to the records the temporary table can be undone.
    */
   private boolean undoable = false;

   /** Table fields by names. */
   private final Map<String, P2JField> fields = new HashMap<String, P2JField>();

   /** Table indexes. */
   private final Map<String, P2JIndex> indexes = new LinkedHashMap<String, P2JIndex>();

   /** Generated DMO buffer class. */
   private Class dmoIface = null;

   /** Handle to default buffer created for the temporary table. */
   private handle defaultBufferHandle = null;

   /** Corresponds ERROR-STRING attribute. **/
   private final character errorString = new character();

   /** Corresponds UNIQUE-ID attribute. **/
   private final integer uniqueId = new integer();

   /** Corresponds ERROR attribute. **/
   private final logical error = new logical();

   /** Corresponds ADM-DATA attribute. **/
   private final character admData = new character();

   /** Persistence class generated for this temporary table. **/
   private PersistentClass persistentClass = null;

   static
   {
      // should go first, see validateDatatype
      datatypesMatchLen.put(DATETIMETZ_TYPE_NAME, 11);
      datatypesMatchLen.put(ParmType.DT.toString(), 8);
      datatypesMatchLen.put(ParmType.INT64.toString(), 5);

      datatypesMatchLen.put(ParmType.INT.toString(), 1);
      datatypesMatchLen.put(ParmType.CHAR.toString(), 1);
      datatypesMatchLen.put(ParmType.DATE.toString(), 2);
      datatypesMatchLen.put(ParmType.DEC.toString(), 2);
      datatypesMatchLen.put(ParmType.HANDLE.toString(), 6);
      datatypesMatchLen.put(ParmType.LOG.toString(), 1);
      datatypesMatchLen.put(ParmType.RAW.toString(), 2);
      datatypesMatchLen.put(ParmType.RECID.toString(), 2);
      datatypesMatchLen.put(ParmType.ROWID.toString(), 3);
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the new temp table, 
    *          if any, or empty string to create the object in the current 
    *          default widget pool.
    */
   public static void create(handle tth, String widgetPool)
   {
      TempTableBuilder builder = new TempTableBuilder();
      builder.uniqueId.assign(UniqueIdGenerator.getUniqueId(UniqueIdGenerator.IdKind.TEMP_TABLE));
      tth.assign(builder);
      // TODO add to widget pool
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the new temp table, 
    *          if any, or empty string to create the object in the current 
    *          default widget pool.
    */
   public static void create(handle tth, character widgetPool)
   {
      if (widgetPool.isUnknown())
      {
         TempTableBuilder.create(tth, "");
      }
      else
      {
         TempTableBuilder.create(tth, widgetPool.getValue());
      }
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    *
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    */
   public static void create(handle tth)
   {
      TempTableBuilder.create(tth, "");
   }

   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handler assigned to a table buffer to create the copy.
    *
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical createTableLike(handle hbuf)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    *
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical createTableLike(String tableName)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character name, character type)
   {
      return addNewField(name, type, null, null, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character name, character type, integer extent)
   {
      return addNewField(name, type, extent, null, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character name, character type, integer extent, character format)
   {
      return addNewField(name, type, extent, format, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial)
   {
      return addNewField(name, type, extent, format, initial, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label)
   {
      return addNewField(name, type, extent, format, initial, label, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * <p>
    * When a parameter is set to null, it means it was not specified by the business logic.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    * @param   columnLabel
    *          The label of the column associated with the defined field
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label,
                              character    columnLabel)
   {
      String fieldName = name.getValue();
      String typeName = type.getValue();

      if (fieldName == null || typeName == null)
      {
         ErrorManager.displayError(9050,
                                   "Unable to evaluate arguments for TEMP-TABLE method",
                                   false);
         return new logical(false);
      }

      if (fields.containsKey(fieldName))
      {
         // we have the field with the same name
         return new logical(false);
      }

      // validate data type
      ParmType parmType = validateDatatype(typeName);
      if (parmType == null)
         return new logical(false);

      String formatStr = format != null ? format.getValue() : null;
      if (formatStr != null)
      {
         try
         {
            // validate format
            switch (parmType)
            {
               case INT:
               case DEC:
               case INT64:
               case HANDLE:
               case RECID:
                  new NumberFormat(formatStr);
                  break;

               case DATE:
                  new date().toString(formatStr);
                  break;

               // TODO handle datetime[-tz] format validation
            }
         }
         catch (DisplayFormatParsingException e)
         {
            return new logical(false);
         }
      }

      String labelStr = label != null ? label.getValue() : null;
      if (labelStr == null)
         labelStr = fieldName;

      String columnLabelStr = columnLabel != null ? columnLabel.getValue() : null;
      if (columnLabelStr == null)
         columnLabelStr = labelStr;

      P2JField field = new P2JField(fieldName,
                                    parmType,
                                    extent != null ? extent.getValue() : 0,
                                    formatStr,
                                    initial,
                                    labelStr,
                                    columnLabelStr);
      fields.put(fieldName, field);
      return new logical(true);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name)
   {
      return addNewIndex(name, null, null, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name, logical unique)
   {
      return addNewIndex(name, unique, null, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name, logical unique, logical primary)
   {
      return addNewIndex(name, unique, primary, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    * <p>
    * If a parameter is null, it means it was not set.
    * 
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * @param   word
    *          TRUE if this is a word index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name, logical unique, logical primary, logical word)
   {
      String indexName = name.getValue();
      if(indexName == null)
      {
         ErrorManager.displayError(9050,
                                   "Unable to evaluate arguments for TEMP-TABLE method",
                                   false);
         return new logical(false);
      }

      if (indexes.containsKey(indexName))
      {
         ErrorManager.displayError(3390,
                             String.format("Duplicate index name %s for temp-table", indexName),
                             false);
         return new logical(false);
      }

      P2JIndex index = new P2JIndex(getName().getValue(),
                                    indexName,
                                    unique != null && unique.getValue(),
                                    primary != null && primary.getValue(),
                                    word != null && word.getValue());
      indexes.put(indexName, index);
      return new logical(true);
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(String name, String source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(character name, character source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(character name, handle source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field, String ascending)
   {
      return addFieldToIndex(new character(indexName), 
                             new character(field),
                             new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field, character ascending)
   {
      return addFieldToIndex(new character(indexName), new character(field), ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, String field, String ascending)
   {
      return addFieldToIndex(indexName, new character(field), new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, String field, character ascending)
   {
      return addFieldToIndex(indexName, new character(field), ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, character field, String ascending)
   {
      return addFieldToIndex(new character(indexName), field, ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, character field, character ascending)
   {
      return addFieldToIndex(new character(indexName), field, ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, character field, String ascending)
   {
      return addFieldToIndex(indexName, field, new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, character field, character ascending)
   {
      String idxName = indexName.getValue();
      String fieldName = field.getValue();
      if(idxName == null || fieldName == null)
      {
         ErrorManager.displayError(9050,
                                   "Unable to evaluate arguments for TEMP-TABLE method",
                                   false);
         return new logical(false);
      }

      String ascVal = ascending.getValue();
      boolean asc = (ascVal == null);

      if (!asc)
      {
         ascVal = ascVal.toLowerCase();
         asc = ascVal.startsWith("asc");

         if (!asc)
         {
            if (!ascVal.startsWith("desc"))
            {
               ErrorManager.displayError(9062,
                             "Third argument to ADD-INDEX-FIELD must be DESC or ASC, in quotes",
                             false);
               return new logical(false);
            }
         }
      }

      P2JField p2jField = fields.get(fieldName);

      // field with this name exist...
      boolean invalidField = (p2jField == null);

      // ...and wasn't already added to the index
      P2JIndex index = indexes.get(idxName);
      if (!invalidField)
      {
         Iterator<P2JIndexComponent> iter = index.components();
         while (iter.hasNext())
         {
            P2JIndexComponent component = iter.next();
            if (component.getName().equals(fieldName))
            {
               invalidField = true;
               break;
            }
         }
      }

      if (invalidField)
      {
         ErrorManager.displayError(9063, String.format(
               "Attempt to add duplicate or inappropriate word index field %s to index %s",
               fieldName,
               idxName),
               false);
         return new logical(false);
      }

      if (p2jField.getExtent() > 1)
      {
         ErrorManager.displayError(370, "Array fields cannot be Index-Fields");
         return new logical(false);
      }

      index.addComponent(fieldName, !asc, false, false, false);
      return new logical(true);
   }

   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(String indexName, String field)
   {
      return addFieldToIndex(new character(indexName), 
                             new character(field), new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(character indexName, String field)
   {
      return addFieldToIndex(indexName, new character(field), new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(String indexName, character field)
   {
      return addFieldToIndex(new character(indexName), field, new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(character indexName, character field)
   {
      return addFieldToIndex(indexName, field, new character("asc"));
   }

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addIndexLike(String name, String sourceName, String dbTable)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addIndexLike(character name,
                               character sourceName,
                               character dbTable)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Returns <code>true</code> if this temp-table is in prepared state (after a successful call
    * to {@link #tempTablePrepare}).
    * This method implements the read-only PREPARED attribute from 4GL.
    *
    * @return  <code>true</code> if this table is prepared.
    */
   public logical prepared()
   {
      return new logical(_prepared());
   }

   /**
    * Returns <code>true</code> if this temp-table is in prepared state (after a successful call
    * to {@link #tempTablePrepare}).
    * This method implements the read-only PREPARED attribute from 4GL.
    *
    * @return  <code>true</code> if this table is prepared.
    */
   public boolean _prepared()
   {
      return !super.getName().isUnknown();
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode. 
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful 
    */
   public logical copyTempTable(handle    other,
                                logical   append,
                                logical   replace,
                                logical   looseCopy,
                                character prefix)
   {
      // TODO implement actua logic
      return new logical(true);
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode. 
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful 
    */
   public logical copyTempTable(handle  other,
                                boolean append,
                                boolean replace,
                                boolean looseCopy,
                                String  prefix)
   {
      // TODO implement actua logic
      return new logical(true);
   }
   /**
    * Sets the UNDO attribute of 4GL temp-table.
    *
    * @param   undo
    *          The new value for the attribute.
    */
   public void setCanUndo(logical undo)
   {
      if (undo.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute UNDO on TEMP-TABLE widget",
               false);
         return;
      }

      this.undoable = undo.booleanValue();

      // TODO UNDO attribute can changed (ant will take effect!) after the table has been prepared
   }

   /** 
    * Obtain the current UNDO attribute of 4GL temp-table.
    * 
    * @return  true if the the temp-table can undo.
    */
   public logical canUndo()
   {
      return new logical(undoable);
   }

   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param   errStr
    *          The new value for the ERROR-STRING attribute.
    */
   public void changeErrorString(character errStr)
   {
      if (!_prepared())
      {
         ErrorManager.displayError(12346,
           "ERROR-STRING can only be used with a PREPARED temp-table that is not a BEFORE-TABLE",
           false);
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR-STRING in widget of type TEMP-TABLE",
               false);
         return;
      }

      this.errorString.assign(errStr);
   }

   /** 
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    * 
    * @return  the current ERROR-STRING of the temp-table
    */
   public character errorString()
   {
      return errorString;
   }

   /**
    * Checks  whether the widget is dynamic or static.
    * 
    * @return  Always returns true for dynamically created temp-tables. 
    */
   public logical dynamic()
   {
      return new logical(true);
   }
   
   /**
    * Checks whether this temp-table contains any records.
    * 
    * @return  true if this temp-table has records.
    */
   public logical hasRecords()
   {
      if (!_prepared())
         return new logical(false);

      Persistence persistence = PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);
      String sql = String.format("select 1 from %s", persistentClass.getTable().getName());
      try
      {
         ScrollableResults res = persistence.executeSQLQuery(sql);
         if (res.first())
            return new logical(true);
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
         return new logical(false);
      }

      // check for uncommitted records in buffers
      if (handle.unwrapBuffer(defaultBufferHandle)._available())
         return new logical(true);

      // TODO check in other buffers

      return new logical(false);
   }

   /**
    * Returns the temp-table to it's clear state, removing all defined fields, indexes, all
    * records in the table and all corresponding buffers.
    * 
    * @return  true if operation is successful
    */
   public logical clear()
   {
      if (fields.size() == 0)
         return new logical(true);

      if (undoable && _prepared() && hasRecords().getValue())
      {
         ErrorManager.displayError(9085,
               "CLEAR method may not be called on a static temp-table object, nor on a dynamic " +
               "one with records; use buffer:EMPTY-TEMP-TABLE",
               false);
         return new logical(false);
      }

      return new logical(forceClear());
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical tempTablePrepare(String name)
   {
      return tempTablePrepare(new character(name));
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical tempTablePrepare(character name)
   {
      if (_prepared())
      {
         ErrorManager.displayError(9043, "TEMP-TABLE is already prepared--ignoring", false);
         return new logical(false);
      }

      String tableName = name.getValue();
      if (tableName == null)
      {
         ErrorManager.displayError(9046, "Unable to PREPARE TEMP-TABLE buffer object", false);
         return new logical(false);
      }

      if (fields.size() == 0)
      {
         ErrorManager.displayError(9033,
                                   "TEMP-TABLE object must have at least one field before " +
                                      "TEMP-TABLE-PREPARE can be called.",
                                   false);
         displayUnableToPrepareTableError();
         return new logical(false);
      }

      int nameLength = tableName.length();
      if (nameLength == 0 || nameLength > 32)
      {
         ErrorManager.displayError(1700,
                                   "Identifier was left blank or is more than 32 characters",
                                   false);
         displayUnableToPrepareTableError();
         return new logical(false);
      }

      if (!Character.isLetter(tableName.charAt(0)))
      {
         ErrorManager.displayError(257,
                     String.format("The first character of %s must be alphabetic", tableName));
         displayUnableToPrepareTableError();
         return new logical(false);
      }

      for (int i = 1; i < nameLength; i++)
      {
         char c = tableName.charAt(i);
         if (!Character.isLetter(c) && !Character.isDigit(c) && !allowedSpecialChars.contains(c))
         {
            ErrorManager.displayError(274,
                     String.format("The character %s is not permitted in name %s", c, tableName));
            displayUnableToPrepareTableError();
            return new logical(false);
         }
      }

      boolean res = DynamicTablesHelper.getInstance().generateDynamicDMO(this);
      if (res)
      {
         createDefaultBuffer();
         super.setName(name); // also makes table "prepared"
      }

      return new logical(res);
   }
   
   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   buffh
    *          A handle to a buffer handle from which to copy the field definitions.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle buffh)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   buffh
    *          A handle to a buffer handle from which to copy the field definitions.
    * @param   except
    *          A list of fields names to be excluded in from the copy process.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle buffh, String... except)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A list of fields names to be excluded in from the copy process.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName, String... except)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Obtain the handle to the buffer that was created by default for this 
    * temporary table. Every dynamic temp-table is created with at 
    * least one buffer. This buffer's object handle is returned by this
    * method. It cannot be called until the table has been prepared using 
    * <code>tempTablePrepare()</code> method.
    * This method is the P2J equivalent of <code>DEFAULT-BUFFER-HANDLE</code>
    * attribute.
    * 
    * @return  The default buffer handle for the temporary table.
    */
   public handle defaultBufferHandle()
   {
      if (!_prepared())
      {
         displayNotPrepared();
         return new handle();
      }

      return defaultBufferHandle;
   }

   /**
    * Set table NAME attribute. Note that it can be changed only after the table has been
    * prepared.
    *
    * @param name
    *        New table name.
    */
   public void setName(character name)
   {
      if (name.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute NAME on TEMP-TABLE widget",
               false);
         return;
      }

      if (!_prepared())
      {
         displayNotPrepared();
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute NAME in widget of type TEMP-TABLE",
               false);
         return;
      }

      super.setName(name);
   }

   /**
    * Get table NAME attribute.
    *
    * @return  name
    *          Table NAME attribute value. If the table hasn't been prepared, unknown value is
    *          returned.
    */
   public character getName()
   {
      character name = super.getName();
      if (name.isUnknown())
      {
         displayNotPrepared();
      }

      return name;
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public logical isError()
   {
      if (!_prepared())
      {
         displayNotPrepared();
      }

      return error;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      setError(new logical(value));
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute ERROR on TEMP-TABLE widget",
               false);
         return;
      }

      if (!_prepared())
      {
         displayNotPrepared();
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR in widget of type TEMP-TABLE",
               false);
         return;
      }

      error.assign(value);
   }

   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NAME attribute only in read-only mode.
    * 
    * @return   Always <code>false</code>, as NAME attribute is writable.
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return false;
   }

   /**
    * Gets the the unique ID number associated to this object by the
    * underlying system. Not the same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   public integer getUniqueID()
   {
      return uniqueId;
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return true;
   }

   /**
    * Worker to be implemented by each resource.  Called by {@link #delete()}.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   protected boolean resourceDelete()
   {
      return forceClear();
   }
   
   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   public character getADMData()
   {
      return admData;
   }

    /**
     * Set the value of the ADM-DATA attribute.
     * 
     * @param    value
     *           The new value.
     */
   public void setADMData(String value)
   {
      setADMData(new character(value));
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   public void setADMData(character value)
   {
      admData.assign(value);
   }

   /**
    * Get the collection of table fields.
    *
    * @return  the collection of table fields.
    */
   public Collection<P2JField> getFields()
   {
      return fields.values();
   }

   /**
    * Get table indexes.
    *
    * @return  the map of table indexes keyed by their names.
    */
   public Map<String, P2JIndex> getIndexes()
   {
      return indexes;
   }

   /**
    * Set DMO interface class which corresponds the table.
    *
    * @param dmoIface
    *        DMO interface class which corresponds the table.
    */
   void setDmoIface(Class dmoIface)
   {
      this.dmoIface = dmoIface;
   }

   /**
    * Set persistent class which corresponds the table.
    *
    * @param persistentClass
    *        Persistent class which corresponds the table.
    */
   void setPersistentClass(PersistentClass persistentClass)
   {
      this.persistentClass = persistentClass;
   }

   /**
    * Create default buffer for the table.
    */
   private void createDefaultBuffer()
   {
      if (defaultBufferHandle == null)
      {
         Temporary buffer =
               DynamicTablesHelper.getInstance().createDynamicBuffer(dmoIface,
                                                            persistentClass.getTable().getName(),
                                                            undoable);
         defaultBufferHandle = new handle();
         defaultBufferHandle.assign(buffer);
      }
   }

   /**
    * Validates data type name which was passed to ADD-NEW-FIELD function.
    *
    * @param  datatype
    *         Data type name.
    *
    * @return corresponding {@link ParmType} or <code>null</code> if validation has failed.
    */
   private ParmType validateDatatype(String datatype)
   {
      // get first word
      int spaceIdx = datatype.indexOf(' ');
      String compareStr = spaceIdx >= 0 ? datatype.substring(0, spaceIdx) : datatype;

      if (compareStr.length() > 0)
      {
         compareStr = compareStr.toLowerCase();

         for (String type : datatypesMatchLen.keySet())
         {
            if (type.startsWith(compareStr))
            {
               if (compareStr.length() >= datatypesMatchLen.get(type))
               {
                  ParmType res = ParmType.fromString(type);
                  if (res == null && DATETIMETZ_TYPE_NAME.equals(type))
                  {
                     res = ParmType.DTTZ;
                  }
                  return res;
               }
            }
         }
      }

      ErrorManager.displayError(9051,
                            String.format("Illegal datatype given in ADD-NEW-FIELD %s", datatype),
                            false);
      return null;
   }

   /**
    * Display "Unable to prepare TEMP-TABLE fields" error message.
    */
   private void displayUnableToPrepareTableError()
   {
      ErrorManager.displayError(9044, "Unable to prepare TEMP-TABLE fields", false);
   }

   /**
    * Display "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc"
    * error message.
    */
   private void displayNotPrepared()
   {
      ErrorManager.displayError(9040,
               "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc",
               false);
   }

   /**
    * Unconditionally delete all buffers for this table, drop DMO definitions and temporary table,
    * clear TEMP-TABLE data structures.
    *
    * @return <code>true</code> if operation was successful.
    */
   private boolean forceClear()
   {
      DynamicTablesHelper helper = DynamicTablesHelper.getInstance();

      // delete all records
      if (defaultBufferHandle != null)
      {
         Buffer buffer = handle.unwrapBuffer(defaultBufferHandle);
         buffer.release();
         ((Deletable) buffer).delete();
         defaultBufferHandle = null;

         // TODO delete all other buffers for this table
      }

      helper.dropDynamicDMO(dmoIface, true);

      // keep UNIQUE-ID attribute
      undoable = true; // however default value is "false"
      fields.clear();
      indexes.clear();
      errorString().setUnknown();
      error.setUnknown();
      admData.setUnknown();
      persistentClass = null;
      dmoIface = null;
      super.setName(new character());

      return true;
   }
}
